import hashlib
import os
import shutil
import typing as t
from pathlib import Path, WindowsPath

from loguru import logger

from pipelime.remotes.base import BaseRemote, NetlocData


class SharedFolderRemote(BaseRemote):
    _PL_FOLDER_ = ".pl"
    _TAGS_FILE_ = "tags.json"
    _HASH_FN_KEY_ = "__HASH_FN__"
    _DEFAULT_HASH_FN_ = "blake2b"

    def __init__(self, netloc_data: NetlocData):
        """Filesystem-based remote.

        Args:
            netloc_data (NetlocData): the network data info.
        """
        if netloc_data.host == "localhost" or netloc_data.host == "127.0.0.1":
            netloc_data.host = ""
        super().__init__(netloc_data)

    def _maybe_create_root(self, target_base_path: Path):
        if not target_base_path.exists():  # pragma: no branch
            logger.info(f"Creating folder tree '{target_base_path}'.")
            target_base_path.mkdir(parents=True, exist_ok=True)
        pldir = target_base_path / self._PL_FOLDER_
        if not pldir.is_dir():  # pragma: no branch
            pldir.mkdir(parents=True, exist_ok=True)

    def _get_hash_fn(self, target_base_path: str) -> t.Any:
        if self.is_valid:  # pragma: no branch
            try:
                import json

                target_root = self._make_file_path(target_base_path, "")
                self._maybe_create_root(target_root)

                tags = {}
                try:
                    with open(
                        target_root / self._PL_FOLDER_ / self._TAGS_FILE_, "r"
                    ) as jtags:
                        tags = json.load(jtags)
                except Exception:
                    pass

                hash_fn_name = tags.get(self._HASH_FN_KEY_)

                # try-get
                if isinstance(hash_fn_name, str) and len(hash_fn_name) > 0:
                    try:
                        hash_fn = getattr(hashlib, hash_fn_name)
                        return hash_fn()
                    except AttributeError:
                        pass

                tags[self._HASH_FN_KEY_] = self._DEFAULT_HASH_FN_
                with open(
                    target_root / self._PL_FOLDER_ / self._TAGS_FILE_, "w"
                ) as jtags:
                    json.dump(tags, jtags)

                hash_fn = getattr(hashlib, self._DEFAULT_HASH_FN_)
                return hash_fn()
            except Exception as exc:
                logger.debug(str(exc))
        return None

    def target_exists(self, target_base_path: str, target_name: str) -> bool:
        if self.is_valid:  # pragma: no branch
            return self._make_file_path(target_base_path, target_name).exists()
        return False

    def _upload(
        self,
        local_stream: t.BinaryIO,
        local_stream_size: int,
        target_base_path: str,
        target_name: str,
    ) -> bool:
        if self.is_valid:  # pragma: no branch
            try:
                target_full_path = self._make_file_path(target_base_path, target_name)
                self._maybe_create_root(target_full_path.parent)

                with target_full_path.open("wb") as target:
                    shutil.copyfileobj(local_stream, target)

                return True
            except Exception as exc:
                logger.debug(str(exc))
                return False

        return False

    def _download(
        self,
        local_stream: t.BinaryIO,
        source_base_path: str,
        source_name: str,
        source_offset: int,
    ) -> bool:
        if self.is_valid:  # pragma: no branch
            try:
                source_full_path = self._make_file_path(source_base_path, source_name)
                if not source_full_path.is_file():
                    logger.debug(f"File '{source_full_path}' does not exist.")
                    return False

                with source_full_path.open("rb") as source:
                    source.seek(source_offset)
                    shutil.copyfileobj(source, local_stream)

                return True
            except Exception as exc:
                logger.debug(str(exc))
                return False

        return False

    def _make_file_path(self, file_path: str, file_name: str) -> Path:
        full_path = Path(file_path) / Path(file_name)
        if self.netloc:
            return Path(
                "{0}{0}{1}{0}{2}".format(os.path.sep, self.netloc, str(full_path))
            )
        elif isinstance(full_path, WindowsPath):
            return full_path
        else:
            return Path(os.path.sep + str(full_path))

    @classmethod
    def scheme(cls) -> str:
        return "file"

    @property
    def is_valid(self) -> bool:
        return True



from pipelime.piper.progress.estimator.base import Estimator
from pipelime.piper.progress.estimator.naive import NaiveEstimator


class EstimatorFactory:
    """Factory for ``Estimator`` s"""

    DEFAULT_ESTIMATOR_TYPE = "NAIVE"

    CLASS_MAP = {
        "NAIVE": NaiveEstimator,
    }

    @classmethod
    def get_estimator(cls, type_: str = DEFAULT_ESTIMATOR_TYPE, **kwargs) -> Estimator:
        return cls.CLASS_MAP[type_](**kwargs)



from pydantic import Field

import pipelime.commands.interfaces as pl_interfaces
from pipelime.piper import PipelimeCommand, PiperPortType


class ToyDatasetCommand(PipelimeCommand, title="toy_dataset"):
    """A generator of random data (images, masks, objects...)."""

    toy: pl_interfaces.ToyDatasetInterface = Field(
        ..., alias="t", description="Toy dataset creation options."
    )

    output: pl_interfaces.OutputDatasetInterface = (
        pl_interfaces.OutputDatasetInterface.pyd_field(
            alias="o", piper_port=PiperPortType.OUTPUT
        )
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    def run(self):
        seq = self.toy.create_dataset_generator()
        seq = self.output.append_writer(seq)
        self.grabber.grab_all(
            seq,
            grab_context_manager=self.output.serialization_cm(),
            keep_order=False,
            parent_cmd=self,
            track_message=f"Writing toy dataset ({len(seq)} samples)",
        )



import time
import zmq
from typing import Optional
from threading import Lock

from pipelime.piper.progress.model import ProgressUpdate
from pipelime.piper.progress.tracker.base import TrackCallback


class ZmqTrackCallback(TrackCallback):
    """ZMQ tracker callback"""

    _addr: str
    _socket: Optional[zmq.Socket]

    PROTOTYPES = {}
    LOCK = Lock()

    def __new__(cls, addr: str = "tcp://*:5556"):
        with cls.LOCK:
            proto = cls.PROTOTYPES.get(addr)
            if proto is None:
                proto = super().__new__(cls)
                proto._addr = addr
                proto._socket = None
                cls.PROTOTYPES[addr] = proto
        return proto

    def update(self, prog: ProgressUpdate):
        if self.socket:
            topic = prog.op_info.token
            self.socket.send_multipart([topic.encode(), prog.json().encode()])

            if prog.finished:
                self.socket.close()
                self._socket = None  # free resources

    @property
    def socket(self):
        if not self._socket:
            self._socket = zmq.Context().socket(zmq.PUB)
            self._socket.bind(self._addr)

            # Wait for the socket to be ready...
            # Apparently, this is the only way to do it. I don't know why.
            time.sleep(1)
        return self._socket



from __future__ import annotations

import copy
import re
import typing as t

from pipelime.items import Item, YamlMetadataItem


class SamplePathRegex:
    KEY_PATH_REGEX = re.compile(r"(?<!\\)(?:\\\\)*\.|(\[\d+\])")  # noqa: W605

    @classmethod
    def split(cls, key_path: str) -> t.Tuple[str, str]:
        key = cls.KEY_PATH_REGEX.split(key_path)[0]
        path = key_path[len(key) :]  # noqa: E203
        key = key.replace(r"\\", "\\").replace(r"\.", r".")
        return key, path


class Sample(t.Mapping[str, Item]):
    """A Sample is a mapping from keys to Items. Any modification creates a new instance
    which is a shallow copy of the original Sample, ie, Item instances are shared, but
    the internal data mapping is not.
    """

    _data: t.Mapping[str, Item]

    def __init__(self, data: t.Optional[t.Mapping[str, Item]] = None):
        super().__init__()
        self._data = {} if data is None else data

    def to_schema(self) -> t.Mapping[str, t.Type[Item]]:
        """Returns a mapping of the keys to the item types."""
        return {k: type(v) for k, v in self._data.items()}

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Returns a dictionary of the keys to the item values."""
        return {k: v() for k, v in self.items()}

    def shallow_copy(self) -> Sample:
        """Returns a shallow copy of the Sample, ie, the Item instances are shared."""
        return Sample(copy.copy(self._data))

    def deep_copy(self) -> Sample:
        """Returns a deep copy of the Sample, ie, the Item instances are not shared."""
        return Sample(copy.deepcopy(self._data))

    def set_item(self, key: str, value: Item) -> Sample:
        """Sets a new Item in the Sample.
        If the key already exists, the Item is replaced.
        """
        new_data = dict(self._data)
        new_data[key] = value
        return Sample(new_data)

    def set_value_as(
        self,
        target_key: str,
        reference_key: str,
        value: t.Any,
        shared_item: t.Optional[bool] = None,
    ) -> Sample:
        """Creates a new Item of the same type of the reference Item and with the
        given value. Then sets the new Item in the Sample with the target key.
        """
        ref_item = self._data[reference_key]
        new_data = dict(self._data)
        new_data[target_key] = ref_item.make_new(
            value, shared=ref_item.is_shared if shared_item is None else shared_item
        )
        return Sample(new_data)

    def set_value(self, key: str, value: t.Any) -> Sample:
        """Sets a new value for the Item with the given key in the Sample."""
        return self.set_value_as(key, key, value)

    def deep_set(
        self,
        key_path: str,
        value: t.Any,
        default_item_type: t.Type[Item] = YamlMetadataItem,
    ) -> Sample:
        r"""Sets a value of an Item of this Sample through a path similar to
        `pydash.set_`. The path is built by splitting the mapping keys by `.`
        and enclosing list indexes within `[]`. Use `\` to escape the `.` character.

        Example::

            sample = Sample({"first": JsonMetadataItem({r"j.names\": ["Jo", "Jane"]})})
            sample.deep_set(r"first.j\.names\\[1]", "Jane Doe")
        """
        import pydash as py_

        key, path = SamplePathRegex.split(key_path)
        x = self if key in self._data else self.set_item(key, default_item_type({}))
        if not path:
            return x.set_value(key, value)

        new_value = copy.deepcopy(x[key]())
        py_.set_(new_value, path, value)
        return x.set_value(key, new_value)

    def deep_get(self, key_path: str, default: t.Any = None) -> t.Any:
        r"""Gets a value from the Sample through a key path similar to `pydash.get`.
        The path is built by splitting the mapping keys by `.` and enclosing list
        indexes within `[]`. Use `\` to escape the `.` character.

        Example::

            sample = Sample({"first": JsonMetadataItem({r"j.names\": ["Jo", "Jane"]})})
            jane = sample.deep_get(r"first.j\.names\\[1]")
        """
        import pydash as py_

        return py_.get(self.direct_access(), key_path, default)

    def match(self, query: str) -> bool:
        """Match the Sample against a query
        (cfr. https://github.com/cyberlis/dictquery).
        """
        import dictquery as dq

        return dq.match(self.direct_access(), query)

    def direct_access(self) -> t.Mapping[str, t.Any]:
        """Returns a mapping of the keys to the item values, ie, you directly get the
        value of the items without having to `__call__()` them."""

        class _DirectAccess(t.Mapping[str, t.Any]):
            def __init__(self, data: t.Mapping[str, Item]):
                self._data = data

            def __getitem__(self, key: str) -> t.Any:
                return self._data[key]()

            def __iter__(self) -> t.Iterator[str]:
                return iter(self._data)

            def __len__(self) -> int:
                return len(self._data)

            def __contains__(self, key: str) -> bool:
                return key in self._data

        return _DirectAccess(self._data)

    def change_key(self, old_key: str, new_key: str, delete_old_key: bool) -> Sample:
        """Creates a new key in the Sample sharing the same Item as the old key.
        It does nothing if `old_key` does not exist or `new_key` already exists.
        """
        if new_key not in self._data and old_key in self._data:
            new_data = dict(self._data)
            new_data[new_key] = new_data[old_key]
            if delete_old_key:
                del new_data[old_key]
            return Sample(new_data)
        return self

    def duplicate_key(self, reference_key: str, new_key: str) -> Sample:
        """Duplicates `reference_key` in the Sample as `new_key`.
        NB: `reference_key` must exist, while `new_key` must not exist.
        """
        return self.change_key(reference_key, new_key, False)

    def rename_key(self, old_key: str, new_key: str) -> Sample:
        """Renames `old_key` in the Sample as `new_key`.
        NB: `old_key` must exist, while `new_key` must not exist.
        """
        return self.change_key(old_key, new_key, True)

    def remove_keys(self, *key_to_remove: str) -> Sample:
        """Removes the given keys from the Sample."""
        return Sample({k: v for k, v in self._data.items() if k not in key_to_remove})

    def extract_keys(self, *keys_to_keep: str) -> Sample:
        """Creates a new Sample with only the given keys."""
        return Sample({k: v for k, v in self._data.items() if k in keys_to_keep})

    def merge(self, other: Sample) -> Sample:
        """Merges the Items in the given Sample."""
        return Sample({**self._data, **other._data})

    def update(self, other: Sample) -> Sample:
        """Same as `merge`."""
        return self.merge(other)

    def zip(self, other: Sample) -> Sample:
        """Same as `merge`."""
        return self.merge(other)

    def __add__(self, other: Sample) -> Sample:
        """Same as `merge`."""
        return self.merge(other)

    def __sub__(self, other: Sample) -> Sample:
        """Removes other's keys from this Sample."""
        return self.remove_keys(*other.keys())

    def __or__(self, other: Sample) -> Sample:
        """Same as `merge`."""
        return self.merge(other)

    def __and__(self, other: Sample) -> Sample:
        """Creates a new Sample with only the keys that are in both Samples."""
        return self.extract_keys(*other.keys())

    def __xor__(self, other: Sample) -> Sample:
        """Creates a new Sample with only the keys that are not in both Samples."""
        return (self | other) - (self & other)

    def __getitem__(self, key: str) -> Item:
        return self._data[key]

    def __iter__(self) -> t.Iterator[str]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def __repr__(self) -> str:
        return f"{self.__class__}{repr(self._data)}"

    def __str__(self) -> str:
        return "\n".join([f"[{k}] {str(v)}" for k, v in self._data.items()])

    def __pl_pretty__(self) -> t.Any:
        from rich.panel import Panel
        from rich.columns import Columns
        from rich import box

        panels = [
            Panel.fit(v.__pl_pretty__(), title=k, title_align="left")
            for k, v in self._data.items()
        ]
        return Panel.fit(
            Columns(panels), title="Sample", title_align="center", box=box.HORIZONTALS
        )



import hashlib
import typing as t
from abc import ABCMeta, abstractmethod
from pathlib import Path
from urllib.parse import ParseResult, unquote_plus, urlparse

from loguru import logger


class NetlocData:
    __slots__ = "host", "user", "password", "init_args"

    host: str
    user: str
    password: str
    init_args: t.Dict[str, t.Union[int, float, str, bool, None]]

    def __init__(self, url: ParseResult):
        user_pwd, _, self.host = url.netloc.rpartition("@")
        self.user, _, self.password = user_pwd.partition(":")

        self.init_args = {
            kw.split("=", 1)[0]: NetlocData._convert_val(kw.split("=", 1)[1])
            for kw in url.query.split(":")
            if len(kw) >= 3 and "=" in kw
        }

    @staticmethod
    def _convert_val(val: str):
        if val.lower() == "true":
            return True
        if val.lower() == "false":
            return False
        if val.lower() in ("none", "null"):
            return None
        try:
            num = int(val)
            return num
        except ValueError:
            pass
        try:
            num = float(val)
            return num
        except ValueError:
            pass
        return val


class RemoteRegister(ABCMeta):
    """A factory class managing a single remote instance
    for any (scheme, netloc) pair.
    """

    REMOTE_CLASSES: t.Dict[str, t.Type["BaseRemote"]] = {}
    REMOTE_INSTANCES: t.Dict[t.Tuple[str, str], "BaseRemote"] = {}

    def __init__(cls, name, bases, dct):
        cls.REMOTE_CLASSES[cls.scheme()] = cls  # type: ignore
        super().__init__(name, bases, dct)

    @classmethod
    def get_instance(
        cls, scheme: str, netloc_data: NetlocData
    ) -> t.Optional["BaseRemote"]:
        """Return a remote instance for this scheme and netloc.

        Args:
          scheme (str): the protocol name, eg, 's3' or 'file'.
          netloc_data (NetlocData): the netloc data info, ie, the host address and
            user account.
        """
        remote_instance = cls.REMOTE_INSTANCES.get((scheme, netloc_data.host))
        if remote_instance is None:
            remote_class = cls.REMOTE_CLASSES.get(scheme)
            if remote_class is not None:
                remote_instance = remote_class(netloc_data)
                cls.REMOTE_INSTANCES[(scheme, netloc_data.host)] = remote_instance
            else:
                logger.warning(f"Unknown remote scheme '{scheme}'.")  # pragma: no cover
        return remote_instance


def make_remote_url(
    *,
    scheme: str = "",
    user: str = "",
    password: str = "",
    host: str = "",
    port: t.Optional[int] = None,
    bucket: t.Union[str, Path] = "",
    **init_args,
) -> ParseResult:
    # NB: we want to normalize the path as "/path/to/loc" or "/c:/path/to/loc"
    if user:
        user_pwd = user
        if password:
            user_pwd += f":{password}"
        user_pwd += "@"
    else:
        user_pwd = ""
    host_port = host + (f":{port}" if port else "")
    return urlparse(
        ParseResult(
            scheme=scheme,
            netloc=f"{user_pwd}{host_port}",
            path=Path(bucket).as_posix(),
            params="",
            query=":".join(f"{k}={v}" for k, v in init_args.items()),
            fragment="",
        ).geturl()
    )


def create_remote(url: ParseResult) -> t.Optional["BaseRemote"]:
    """Return a remote instance for this scheme and netloc.

    Args:
      url (ParseResult): the url describing the remote as
        `<scheme>://[user[:pwd]@]<netloc>/<any_path>[?<init-kw>=<init-val>:<init-kw>=<init-val>...]`

    Returns:
      Optional[BaseRemote]: the remote instance or None if the scheme is unknown.
    """
    return RemoteRegister.get_instance(url.scheme, NetlocData(url))


def paths_from_url(
    url: ParseResult,
) -> t.Union[t.Tuple[str, str], t.Tuple[None, None]]:
    """Extract base path and target name from the path component of a given url.

    Args:
      url (ParseResult): the url to split.

    Returns:
      Union[Tuple[str, str], Tuple[None, None]]: the base path and the target name.
    """
    if len(url.path) > 1:
        unquoted = unquote_plus(url.path)
        file_full_path = Path(unquoted[1:] if unquoted.startswith("/") else unquoted)
        if not file_full_path.suffix:
            return str(file_full_path.as_posix()), ""
        return str(file_full_path.parent.as_posix()), str(file_full_path.name)
    else:
        return None, None  # pragma: no cover


class BaseRemote(metaclass=RemoteRegister):  # type: ignore
    """Base class for any remote."""

    def __init__(self, netloc_data: NetlocData):
        """Set the network address.

        Args:
            netloc_data (NetlocData): the network data info.
        """
        self._netloc = netloc_data.host

    def upload_file(
        self, local_file: t.Union[Path, str], target_base_path: str
    ) -> t.Optional[ParseResult]:
        """Upload an existing file to remote storage.

        Args:
          local_file (Union[Path, str]): local file path.
          target_base_path (str): the remote base path, eg, the bucket name.

        Returns:
          Optional[ParseResult]: a url to get back the file.
        """
        local_file = Path(local_file)
        file_size = local_file.stat().st_size
        with open(local_file, "rb") as file_data:
            return self.upload_stream(
                file_data, file_size, target_base_path, "".join(local_file.suffixes)
            )

    def upload_stream(
        self,
        local_stream: t.BinaryIO,
        local_stream_size: int,
        target_base_path: str,
        target_suffix: str,
    ) -> t.Optional[ParseResult]:
        """Upload data from a readable stream.

        Args:
          local_stream (BinaryIO): a binary data stream.
          local_stream_size (int): the byte size to upload.
          target_base_path (str): the remote base path, eg, the bucket name.
          target_suffix (str): the remote file suffix.

        Returns:
          Optional[ParseResult]: a url to get back the file.
        """
        hash_name = self._compute_hash(
            local_stream, self._get_hash_fn(target_base_path)
        )
        target_name = hash_name + target_suffix
        if self.target_exists(target_base_path, target_name) or self._upload(
            local_stream, local_stream_size, target_base_path, target_name
        ):
            return self._make_url(f"{target_base_path}/{target_name}")
        # upload has failed
        return None

    def download_file(
        self,
        local_file: t.Union[Path, str],
        source_base_path: str,
        source_name: str,
    ) -> bool:
        """Download and write data to a local file.

        Args:
          local_file (Union[Path, str]): the file to write to.
          source_base_path (str): the remote base path, eg, the bucket name.
          source_name (str): the remote object name.

        Returns:
          bool: True if no error occurred.
        """
        local_file = Path(local_file)
        source_name_path = Path(source_name)
        if local_file.suffixes != source_name_path.suffixes:
            local_file = local_file.with_suffix("".join(source_name_path.suffixes))
        if local_file.is_file():
            with open(local_file, "rb") as local_stream:
                hash_name = self._compute_hash(
                    local_stream, self._get_hash_fn(source_base_path)
                )
                if (
                    hash_name
                    == source_name_path.name[: -len("".join(source_name_path.suffixes))]
                ):
                    # local file is up-to-date
                    return True

        local_file.parent.mkdir(parents=True, exist_ok=True)
        part_file = local_file.with_suffix(local_file.suffix + ".part")
        offset: int = 0
        try:
            offset = part_file.stat().st_size
        except IOError:
            pass

        ok = False
        with open(part_file, "ab") as part_stream:
            ok = self.download_stream(
                part_stream, source_base_path, source_name, offset
            )

        if ok:
            local_file.unlink(missing_ok=True)
            part_file.rename(local_file)

        return ok

    def download_stream(
        self,
        local_stream: t.BinaryIO,
        source_base_path: str,
        source_name: str,
        source_offset: int = 0,
    ) -> bool:
        """Download and write data to a writable stream.

        Args:
          local_stream (BinaryIO): a writable stream.
          source_base_path (str): the remote base path, eg, the bucket name.
          source_name (str): the remote object name.
          source_offset (int, optional): where to start reading the remote data.
            (Default value = 0)

        Returns:
          bool: True if no error occurred.
        """
        return self._download(
            local_stream, source_base_path, source_name, source_offset
        )

    @abstractmethod
    def target_exists(self, target_base_path: str, target_name: str) -> bool:
        """Check if a remote object exists.

        Args:
          target_base_path (str): the remote base path, eg, the bucket name.
          target_name (str): the remote object name.

        Returns:
          bool: True if <target_base_path>/<target_name> exists on this remote.
        """
        pass

    def _compute_hash(self, stream: t.BinaryIO, hash_fn: t.Any = None) -> str:
        """Compute a hash value based on the content of a readable stream.

        Args:
          stream (BinaryIO): the stream to hash.
          hash_fn (Any, optional): the hash funtion, if None `hashlib.blake2b()` will
            be used  (Default value = None)

        Returns:
          str: the computed hash
        """
        if hash_fn is None:
            hash_fn = hashlib.blake2b()  # pragma: no cover
        b = bytearray(1024 * 1024)
        mv = memoryview(b)
        fpos = stream.tell()
        for n in iter(lambda: stream.readinto(mv), 0):  # type: ignore
            hash_fn.update(mv[:n])
        stream.seek(fpos)
        return hash_fn.hexdigest()

    def _make_url(self, target_full_path: str) -> ParseResult:
        """Make a complete url for this remote path.

        Args:
          target_full_path (str): the remote path.

        Returns:
          ParseResult: the url
        """
        return ParseResult(
            scheme=self.scheme(),
            netloc=self.netloc,
            path=Path(target_full_path).as_posix(),
            params="",
            query="",
            fragment="",
        )

    @abstractmethod
    def _get_hash_fn(self, target_base_path: str) -> t.Any:
        """Return the hash function used for the object in the give remote base path.

        Args:
          target_base_path (str): the remote base path, eg, the bucket name.

        Returns:
          Any: the hash function, eg, `hashlib.blake2b()`.
        """
        pass

    @abstractmethod
    def _upload(
        self,
        local_stream: t.BinaryIO,
        local_stream_size: int,
        target_base_path: str,
        target_name: str,
    ) -> bool:
        """Upload data from a readable stream.

        Args:
          local_stream (BinaryIO): a binary data stream.
          local_stream_size (int): the byte size to upload.
          target_base_path (str): the remote base path, eg, the bucket name.
          target_name (str): the remote object name.

        Returns:
          bool: True if no error occurred.
        """
        pass

    @abstractmethod
    def _download(
        self,
        local_stream: t.BinaryIO,
        source_base_path: str,
        source_name: str,
        source_offset: int,
    ) -> bool:
        """Download and write data to a writable stream.

        Args:
          local_stream (BinaryIO): a writable stream.
          source_base_path (str): the remote base path, eg, the bucket name.
          source_name (str): the remote object name.
          source_offset (int, optional): where to start reading the remote data

        Returns:
          bool: True if no error occurred.
        """
        pass

    @classmethod
    @abstractmethod
    def scheme(cls) -> str:
        """Return the scheme name, eg, 's3' or 'file'.

        Returns:
          str: the scheme name.
        """
        pass

    @property
    def netloc(self) -> str:
        """Return the remote network address.

        Returns:
          str: the remote network address.
        """
        return self._netloc

    @property
    @abstractmethod
    def is_valid(self) -> bool:
        """Check if this remote instance is valid and usable.

        Returns:
          bool: True if this instance is valid.
        """
        pass



from pipelime.sequences.sample import Sample
from pipelime.sequences.samples_sequence import (
    SamplesSequence,
    source_sequence,
    piped_sequence,
)

# import and register all sequence functionals
import pipelime.sequences.sources
import pipelime.sequences.pipes

from pipelime.sequences.grabber import Grabber, grab_all
from pipelime.sequences.utils import build_pipe, DataStream

from pipelime.utils.pydantic_types import SampleValidationInterface, ItemValidationModel



import typing as t
from pathlib import Path
from urllib.parse import ParseResult

import pydantic as pyd

from pipelime.items import Item
from pipelime.stages import SampleStage

if t.TYPE_CHECKING:
    from pipelime.sequences import Sample


class StageUploadToRemote(SampleStage, title="remote-upload"):
    """Uploads the sample to one or more remote servers."""

    remotes: t.Sequence[ParseResult] = pyd.Field(
        ..., description="The remote addresses and buckets."
    )
    keys_to_upload: t.Optional[t.Sequence[str]] = pyd.Field(
        None, description="The keys to upload. If None, all keys are uploaded."
    )

    @pyd.validator("remotes")
    def unique_remotes(cls, v):
        return tuple(set(v))

    def __call__(self, x: "Sample") -> "Sample":
        for k, v in x.items():
            if not self.keys_to_upload or k in self.keys_to_upload:
                v.serialize(*self.remotes)  # type: ignore
        return x


class StageForgetSource(SampleStage, title="forget-source"):
    """Removes data sources, ie, file paths or remotes, from items."""

    always_remove: t.Sequence[t.Union[Path, ParseResult]] = pyd.Field(
        default_factory=list, description="This sources will be removed from any item."
    )
    remove_by_key: t.Mapping[str, t.Sequence[t.Union[Path, ParseResult]]] = pyd.Field(
        default_factory=dict,
        description="Sources to be removed from specific sample keys.",
    )

    def __init__(
        self,
        *always_remove: t.Union[Path, ParseResult],
        **remove_by_key: t.Union[
            t.Union[Path, ParseResult], t.Sequence[t.Union[Path, ParseResult]]
        ]
    ):
        remove_by_key = {
            k: (v,) if isinstance(v, (Path, ParseResult)) else tuple(v)
            for k, v in remove_by_key.items()
        }
        super().__init__(
            always_remove=always_remove, remove_by_key=remove_by_key  # type: ignore
        )

    def __call__(self, x: "Sample") -> "Sample":
        from pipelime.sequences import Sample

        new_data: t.Dict[str, Item] = {}
        for k, v in x.items():
            rm = self.always_remove + self.remove_by_key.get(k, tuple())  # type: ignore
            new_data[k] = v.remove_data_source(*rm)
        return Sample(new_data)



from __future__ import annotations
import multiprocessing
import pydantic as pyd
import typing as t
from enum import Enum, auto

import pipelime.sequences as pls


class ReturnType(Enum):
    NO_RETURN = auto()
    SAMPLE = auto()
    SAMPLE_AND_INDEX = auto()


class Grabber(pyd.BaseModel, extra="forbid", copy_on_model_validation="none"):
    num_workers: int = pyd.Field(
        0,
        description=(
            "The number of processes to spawn. If negative, "
            "the number of (logical) cpu cores is used."
        ),
    )
    prefetch: pyd.PositiveInt = pyd.Field(
        2, description="The number of samples loaded in advanced by each worker."
    )
    keep_order: bool = pyd.Field(
        False, description="Whether to retrieve the samples in the original order."
    )

    def __call__(
        self,
        sequence: pls.SamplesSequence,
        return_type: ReturnType = ReturnType.SAMPLE,
        size: t.Optional[int] = None,
        *,
        worker_init_fn: t.Optional[t.Tuple[t.Callable, t.Sequence]] = None,
    ) -> _GrabContext:
        return _GrabContext(
            self,
            sequence,
            return_type=return_type,
            size=size,
            worker_init_fn=worker_init_fn,
        )


class _GrabWorker:
    def __init__(self, sequence: pls.SamplesSequence):
        self._sequence = sequence

    def _worker_fn_no_return(self, idx) -> None:
        _ = self._sequence[idx]

    def _worker_fn_sample(self, idx) -> pls.Sample:
        return self._sequence[idx]

    def _worker_fn_sample_and_index(self, idx) -> t.Tuple[int, pls.Sample]:
        return idx, self._sequence[idx]


class _GrabContext:
    def __init__(
        self,
        grabber: Grabber,
        sequence: pls.SamplesSequence,
        return_type: ReturnType,
        size: t.Optional[int],
        worker_init_fn: t.Optional[t.Tuple[t.Callable, t.Sequence]],
    ):
        self._grabber = grabber
        self._sequence = sequence
        self._return_type = return_type
        self._size = size
        self._pool = None
        self._worker_init_fn = (None, ()) if worker_init_fn is None else worker_init_fn

    @staticmethod
    def wrk_init(extra_modules, user_init_fn):
        from pipelime.cli.utils import PipelimeSymbolsHelper

        PipelimeSymbolsHelper.set_extra_modules(extra_modules)
        PipelimeSymbolsHelper.import_everything()

        if user_init_fn[0] is not None:
            user_init_fn[0](*user_init_fn[1])

    def __enter__(self):
        from pipelime.cli.utils import PipelimeSymbolsHelper

        if self._grabber.num_workers == 0:
            # SINGLE PROCESS
            self._pool = None
            it = iter(self._sequence)
            if self._worker_init_fn[0] is not None:
                self._worker_init_fn[0](*self._worker_init_fn[1])
            if self._return_type == ReturnType.SAMPLE_AND_INDEX:
                return enumerate(it)
            return it

        # MULTIPLE PROCESSES
        self._pool = multiprocessing.Pool(
            self._grabber.num_workers if self._grabber.num_workers > 0 else None,
            initializer=_GrabContext.wrk_init,
            initargs=(PipelimeSymbolsHelper.extra_modules, self._worker_init_fn),
        )
        runner = self._pool.__enter__()

        worker = _GrabWorker(self._sequence)
        if self._return_type == ReturnType.NO_RETURN:
            fn = worker._worker_fn_no_return
        elif self._return_type == ReturnType.SAMPLE:
            fn = worker._worker_fn_sample
        else:
            fn = worker._worker_fn_sample_and_index

        if self._grabber.keep_order:
            return runner.imap(
                fn,
                range(len(self._sequence) if self._size is None else self._size),
                chunksize=self._grabber.prefetch,
            )
        return runner.imap_unordered(
            fn,
            range(len(self._sequence) if self._size is None else self._size),
            chunksize=self._grabber.prefetch,
        )

    def __exit__(self, exc_type, exc_value, traceback):
        if self._pool is not None:
            self._pool.__exit__(exc_type, exc_value, traceback)  # type: ignore


def grab_all(
    grabber: Grabber,
    sequence: pls.SamplesSequence,
    *,
    track_fn: t.Optional[t.Callable[[t.Iterable], t.Iterable]] = None,
    sample_fn: t.Union[
        t.Callable[[pls.Sample], None], t.Callable[[pls.Sample, int], None], None
    ] = None,
    size: t.Optional[int] = None,
    grab_context_manager: t.Optional[t.ContextManager] = None,
    worker_init_fn: t.Union[t.Callable, t.Tuple[t.Callable, t.Sequence], None] = None,
):
    from inspect import signature, Parameter
    import contextlib

    if track_fn is None:
        track_fn = lambda x: x  # noqa: E731
    if sample_fn is None:
        sample_fn = lambda x: None  # noqa: E731
        return_type = ReturnType.NO_RETURN
    else:
        prms = signature(sample_fn).parameters
        return_type = (
            ReturnType.SAMPLE_AND_INDEX
            if (
                len(prms) > 1
                or next(iter(prms.values())).kind == Parameter.VAR_POSITIONAL
            )
            else ReturnType.SAMPLE
        )
    if grab_context_manager is None:
        grab_context_manager = contextlib.nullcontext()

    if isinstance(worker_init_fn, t.Sequence):
        if len(worker_init_fn) == 1:
            worker_init_fn = (worker_init_fn[0], ())
        elif len(worker_init_fn) != 2:
            raise ValueError(
                "The worker_init_fn argument must be a callable or a tuple of "
                "a callable and its arguments."
            )
        worker_init_fn = tuple(worker_init_fn)
    elif worker_init_fn is not None:
        worker_init_fn = (worker_init_fn, ())

    with grab_context_manager:
        ctx = grabber(
            sequence,
            return_type=return_type,
            size=size,
            worker_init_fn=worker_init_fn,
        )
        if return_type == ReturnType.SAMPLE_AND_INDEX:
            with ctx as gseq:
                for idx, sample in track_fn(gseq):
                    sample_fn(sample, idx)  # type: ignore
        else:
            with ctx as gseq:
                for sample in track_fn(gseq):
                    sample_fn(sample)  # type: ignore



from pathlib import Path
import albumentations as A
import typing as t
import pydantic as pyd

from pipelime.stages import SampleStage

if t.TYPE_CHECKING:
    from pipelime.sequences import Sample


class Transformation(pyd.BaseModel, extra="forbid", copy_on_model_validation="none"):
    """The albumentations transformation defined as python object,
    serialized dict or yaml/json file.
    """

    __root__: t.Dict[str, t.Any]
    _value: t.Union[A.BaseCompose, A.BasicTransform] = pyd.PrivateAttr(None)

    def __init__(self, **data):
        super().__init__(**data)
        self._value = A.from_dict(self.__root__)  # type: ignore

    @property
    def value(self):
        return self._value

    def __str__(self) -> str:
        return str(self.__root__)

    def __repr__(self) -> str:
        return repr(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, Transformation):
            return value
        if isinstance(value, (A.BaseCompose, A.BasicTransform)):
            return Transformation(__root__=A.to_dict(value))
        if isinstance(value, (str, Path)):
            import yaml

            with open(str(value)) as f:
                value = yaml.safe_load(f)
        if isinstance(value, t.Mapping):
            return Transformation(__root__=value)
        raise ValueError(f"{value} is not a valid transformation")


class StageAlbumentations(SampleStage, title="albumentations"):
    """Sample augmentation via Albumentations."""

    transform: Transformation = pyd.Field(...)
    keys_to_targets: t.Mapping[str, str] = pyd.Field(
        ...,
        description=(
            "A mapping from key names to albumentation targets' names. "
            "NB: key names will be replaced with the target name possibly followed "
            "by an index, ie, `image`, `image_0`, `image_1` etc, "
            "`mask`, `mask_0`, `mask_1` etc."
        ),
    )
    output_key_format: str = pyd.Field(
        "*",
        description=(
            "How to format the output keys. Any `*` will be replaced with the "
            "source key, eg, `aug_*_out` on [`image`, `mask`] generates "
            "`aug_image_out` and `aug_mask_out`. If no `*` is found, the string is "
            "suffixed to the source key, ie, `OutName` on `image` gives "
            "`imageOutName`. If empty, the source key will be used as-is."
        ),
    )

    _target_to_keys: t.Dict[str, str] = pyd.PrivateAttr(default_factory=dict)

    @pyd.validator("output_key_format")
    def validate_output_key_format(cls, v):
        if "*" in v:
            return v
        return "*" + v

    def __init__(self, **data):
        super().__init__(**data)

        target_counter: t.Mapping[str, int] = {}
        target_types: t.Mapping[str, str] = {}

        for key, trg in self.keys_to_targets.items():
            counter = target_counter.get(trg, 0)
            trg_name = trg if counter == 0 else f"{trg}_{counter}"
            target_counter[trg] = counter + 1
            target_types[trg_name] = trg
            self._target_to_keys[trg_name] = key

        self.transform.value.add_targets(target_types)

    def __call__(self, x: "Sample") -> "Sample":
        to_transform = {k: x[v]() for k, v in self._target_to_keys.items() if v in x}
        transformed = self.transform.value(**to_transform)
        for k, v in transformed.items():
            x_key = self._target_to_keys[k]
            x = x.set_value_as(self.output_key_format.replace("*", x_key), x_key, v)
        return x



from __future__ import annotations

import typing as t
import functools
import itertools
from pathlib import Path

import typer

from pipelime.cli.subcommands import SubCommands as subc
from pipelime.cli.parser import parse_pipelime_cli, CLIParsingError
from pipelime.cli.utils import (
    PipelimeSymbolsHelper,
    print_command_op_stage_info,
    print_commands_ops_stages_list,
)

if t.TYPE_CHECKING:
    from pipelime.choixe import XConfig


def _complete_yaml(incomplete: str):
    for v in Path(".").glob(f"{incomplete}*.yaml"):
        yield str(v)


def _print_dict(name, data):
    import json

    from pipelime.cli.pretty_print import print_info

    print_info(f"\n{name}:")
    print_info(json.dumps(data, indent=2))


def _update_dispatch(original_value, item):
    if isinstance(item, t.Mapping):
        if isinstance(original_value, t.MutableMapping):
            _dict_update(original_value, item)
            return True
    elif isinstance(item, t.Sequence):
        if isinstance(original_value, t.MutableSequence):
            _list_update(original_value, item)
            return True
    return False


def _list_update(to_be_updated: t.MutableSequence, data: t.Sequence):
    to_be_updated.extend([None] * (len(data) - len(to_be_updated)))  # this is safe
    if data[-1] is None:
        del to_be_updated[len(data) - 1 :]  # noqa: E203
    for idx, item in enumerate(data):
        if item is not None:
            original_value = to_be_updated[idx]
            if _update_dispatch(original_value, item):
                continue
            to_be_updated[idx] = item


def _dict_update(to_be_updated: t.MutableMapping, data: t.Mapping):
    for k, v in data.items():
        if k in to_be_updated:
            original_value = to_be_updated[k]
            if _update_dispatch(original_value, v):
                continue
        to_be_updated[k] = v


def _deep_update_fn(to_be_updated: "XConfig", data: "XConfig", verbose: bool):
    from pipelime.cli.pretty_print import print_info

    if verbose:
        print_info("> Merging XConfigs:")
        print_info(">> Base:")
        print_info(to_be_updated.to_dict(), pretty=True)
        print_info(">> To be merged:")
        print_info(data.to_dict(), pretty=True)
    to_be_updated.deep_update(data, full_merge=True)
    if verbose:
        print_info(">> Result:")
        print_info(to_be_updated.to_dict(), pretty=True)
    return to_be_updated


def _process_cfg_or_die(
    cfg: "XConfig",
    ctx: t.Optional["XConfig"],
    cfg_name: str,
    run_all: t.Optional[bool],
    output: t.Optional[Path],
    exit_on_error: bool,
    verbose: bool,
    print_all: bool,
) -> t.List["XConfig"]:
    from pipelime.cli.pretty_print import print_error, print_info
    from pipelime.choixe.visitors.processor import ChoixeProcessingError
    from pipelime.cli.pretty_print import print_error

    if verbose:
        print_info(f"> Processing {cfg_name}...")

    try:
        effective_configs = (
            [cfg.process(ctx)]
            if run_all is not None and not run_all
            else cfg.process_all(ctx)
        )
    except ChoixeProcessingError as e:
        if exit_on_error:
            print_error(
                f"Invalid {cfg_name}! {e}\nRun with -vv or more to get more info."
            )
            raise typer.Exit(1)
        raise e

    if verbose:
        pls = "s" if len(effective_configs) != 1 else ""
        print_info(f"> Found {len(effective_configs)} {cfg_name}{pls}")

    if len(effective_configs) > 1 and run_all is None:
        if not typer.confirm(
            f"{len(effective_configs)} {cfg_name}s found. "
            "Do you want to keep them all?"
        ):
            effective_configs = effective_configs[:1]

    if print_all:
        for idx, c in enumerate(effective_configs):
            print_info(f">> {cfg_name} {idx}:")
            print_info(c.to_dict(), pretty=True)

    if output is not None:
        zero_fill = len(str(len(effective_configs) - 1))
        for idx, cfg in enumerate(effective_configs):
            filepath = (
                output.with_name(f"{output.name}_{str(idx).zfill(zero_fill)}")
                if len(effective_configs) > 1
                else output
            )
            cfg.save_to(filepath)
    return effective_configs


def _process_all(
    base_cfg: t.List["XConfig"],
    effective_ctx: "XConfig",
    output: t.Optional[Path],
    run_all: t.Optional[bool],
    exit_on_error: bool,
    verbose: int,
):
    from pipelime.cli.pretty_print import print_info
    from pipelime.choixe import XConfig

    if verbose > 1:
        print_info("\nProcessing configurations:")

    # first process with no branch
    effective_configs = [
        _process_cfg_or_die(
            c,
            effective_ctx,
            "configuration",
            False,
            output,
            exit_on_error,
            verbose > 1,
            verbose > 3,
        )
        for c in base_cfg
        if c.to_dict()
    ]
    if effective_configs:
        # effective_configs = functools.reduce(
        #     lambda acc, curr: acc + curr, effective_configs
        # )
        effective_configs = functools.reduce(
            lambda acc, curr: _deep_update_fn(acc, curr, verbose > 3),
            itertools.chain.from_iterable(effective_configs),
        )
    else:
        effective_configs = XConfig()

    if verbose > 2:
        print_info("\nFinal merged configuration:")
        print_info(effective_configs.to_dict(), pretty=True)

    if verbose > 1:
        print_info("\nProcessing branches:")

    # now process the branches, if any, and check the overall merge
    return _process_cfg_or_die(
        effective_configs,
        effective_ctx,
        "configuration",
        run_all,
        output,
        exit_on_error,
        verbose > 1,
        verbose > 3,
    )


class VersionCallback:
    @staticmethod
    def get_version():
        import pipelime

        return pipelime.__version__


def version_callback(value: bool):
    if value:
        print(VersionCallback.get_version())
        raise typer.Exit()


def pl_main(
    ctx: typer.Context,
    config: t.List[Path] = typer.Option(
        None,
        "--config",
        "-c",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help=(
            "One or more yaml/json files with some or "
            "all the arguments required by the command.\n\n"
            "`++opt` or `+opt` command line options update and override them.\n\n "
        ),
        autocompletion=_complete_yaml,
    ),
    context: t.List[Path] = typer.Option(
        None,
        "--context",
        "-x",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help=(
            "One or more yaml/json files with some or all the context parameters. "
            "If `--config` is set and `--context` is not, all files matching "
            "`*context*.[yaml|yml|json]` in the folders of all the "
            "configuration files will be loaded as the context. "
            "Use `--no-ctx-autoload` to disable this behavior.\n\n"
            "`@@opt` or `@opt` command line options update and override them.\n\n"
            "After a `//` token, `++opt` and `+opt` are accepted as well.\n\n "
        ),
        autocompletion=_complete_yaml,
    ),
    ctx_autoload: bool = typer.Option(
        True,
        help=(
            "Enable/disable context auto-loading when "
            "`--config` is set and `--context` is not."
        ),
    ),
    extra_modules: t.List[str] = typer.Option(
        [], "--module", "-m", help="Additional modules to import."
    ),
    run_all: t.Optional[bool] = typer.Option(
        None,
        help=(
            "In case of multiple configurations, run them all or only the first one. "
            "If not specified, user will be notified if multiple configurations "
            "are found and asked on how to proceed."
        ),
    ),
    output: t.Optional[Path] = typer.Option(
        None,
        "--output",
        "-o",
        writable=True,
        resolve_path=True,
        help="Save final processed configuration to json/yaml.",
    ),
    output_ctx: t.Optional[Path] = typer.Option(
        None,
        writable=True,
        resolve_path=True,
        help="Save final processed context to json/yaml.",
    ),
    verbose: int = typer.Option(
        0,
        "--verbose",
        "-v",
        help="Verbose output. Can be specified multiple times.",
        count=True,
    ),
    dry_run: bool = typer.Option(False, "--dry-run", "-d", help="Dry run."),
    command: str = typer.Argument(
        "",
        show_default=False,
        help=(
            (
                "A command, ie, a `command-name`, "
                "a `package.module.ClassName` class path or "
                "a `path/to/module.py:ClassName` uri (use with care).\n\n"
            )
            + subc.get_help()
        ),
        autocompletion=PipelimeSymbolsHelper.complete_name(
            is_cmd=True,
            is_seq_op=False,
            is_stage=False,
            additional_names=subc.get_autocompletions(),
        ),
    ),
    command_args: t.Optional[t.List[str]] = typer.Argument(
        None,
        help=(
            "\b\nExpected command line:\n"
            "- `++opt` and `+opt` are command parameters\n"
            "- `@@opt` and `@opt` are context parameters\n"
            "- after `//` `++opt`, `+opt`, `@@opt`and `@opt` "
            "are always context parameters."
        ),
    ),
    help: bool = typer.Option(
        False, "--help", "-h", show_default=False, help="Show this message and exit."
    ),
    version: t.Optional[bool] = typer.Option(
        None,
        "--version",
        show_default=False,
        help="Show version and exit.",
        callback=version_callback,
        is_eager=True,
    ),
):
    """
    {0} Examples:

    `{1} list` prints a list of the available commands, sequence operators
    and stages.

    `{1} help <cmd-op-stg>` prints informations on a specific command, sequence
    operator or stage.

    `{1} <command> [<args>]` runs a{2} command.

    NB: command (++opt) and context (@@opt) arguments with no value are treated as
    TRUE boolean values. Use `false` or `true` to explicitly set a boolean
    and `none`/`null`/`nul` to enforce `None`.
    """

    PipelimeSymbolsHelper.set_extra_modules(extra_modules)

    if command_args is None:
        command_args = []

    if (
        help
        or command in subc.HELP[0]
        or any([h in command_args for h in subc.HELP[0]])
    ):
        try:
            if command and command not in subc.HELP[0]:
                print_command_op_stage_info(
                    command, show_description=verbose > 0, recursive=verbose > 1
                )
            elif command_args:
                print_command_op_stage_info(
                    command_args[0], show_description=verbose > 0, recursive=verbose > 1
                )
            else:
                print(ctx.get_help())
        except ValueError:
            pass
    elif command in subc.WIZARD[0] or any([w in command_args for w in subc.WIZARD[0]]):
        from pipelime.cli.wizard import Wizard

        if command and command not in subc.WIZARD[0]:
            Wizard.model_cfg_wizard(command)
        Wizard.model_cfg_wizard(command_args[0])
    elif command in subc.LIST[0]:
        print_commands_ops_stages_list(
            verbose > 0,
            show_cmds=True,
            show_ops=True,
            show_stages=True,
            show_description=verbose > 0,
            recursive=verbose > 1,
        )
    elif command in subc.LIST_CMDS[0]:
        print_commands_ops_stages_list(
            verbose > 0,
            show_cmds=True,
            show_ops=False,
            show_stages=False,
            show_description=verbose > 0,
            recursive=verbose > 1,
        )
    elif command in subc.LIST_OPS[0]:
        print_commands_ops_stages_list(
            verbose > 0,
            show_cmds=False,
            show_ops=True,
            show_stages=False,
            show_description=verbose > 0,
            recursive=verbose > 1,
        )
    elif command in subc.LIST_STGS[0]:
        print_commands_ops_stages_list(
            verbose > 0,
            show_cmds=False,
            show_ops=False,
            show_stages=True,
            show_description=verbose > 0,
            recursive=verbose > 1,
        )
    elif command:
        import pipelime.choixe.utils.io as choixe_io
        from pipelime.choixe import XConfig
        from pipelime.cli.pretty_print import print_error, print_info, print_warning

        if config and not context and ctx_autoload:
            context = []
            for c in config:
                for p in c.resolve().parent.glob("*context*.*"):
                    if p.suffix in (".yaml", ".yml", ".json"):
                        context += [p]

        if verbose > 0:

            def _print_file_list(files: t.Sequence[Path], name: str):
                if files:
                    flist = ", ".join(f'"{str(c)}"' for c in files)
                    print_info(
                        f"{name.capitalize()} file"
                        + (f"s: [ {flist} ]" if len(files) > 1 else f": {files[0]}")
                    )
                else:
                    print_info(f"No {name} file")

            _print_file_list(config, "configuration")
            _print_file_list(context, "context")
            print_info(
                f"Other command and context arguments: {command_args}"
                if command_args
                else "No other command or context arguments"
            )

        base_cfg = [choixe_io.load(c) for c in config]
        base_ctx = [choixe_io.load(c) for c in context]

        # process extra args
        try:
            cmdline_cfg, cmdline_ctx = parse_pipelime_cli(command_args)
        except CLIParsingError as e:
            e.rich_print()
            raise typer.Exit(1)

        if verbose > 2:
            _print_dict(
                f"Loaded configuration file{'s' if len(config) > 1 else ''}", base_cfg
            )
            _print_dict(
                f"Loaded context file{'s' if len(context) > 1 else ''}", base_ctx
            )
            _print_dict("Configuration options from command line", cmdline_cfg)
            _print_dict("Context options from command line", cmdline_ctx)

        # keep each config separated to get the right cwd
        base_cfg = [XConfig(data=c, cwd=p.parent) for c, p in zip(base_cfg, config)]
        base_cfg.append(XConfig(data=cmdline_cfg, cwd=Path.cwd()))

        base_ctx = [XConfig(data=c, cwd=p.parent) for c, p in zip(base_ctx, context)]
        base_ctx.append(XConfig(data=cmdline_ctx, cwd=Path.cwd()))

        # process contexts to resolve imports and local loops
        if verbose > 2:
            print_info("\nProcessing context files:")

        def _sum(a: XConfig, b: XConfig) -> XConfig:
            a = XConfig(data=a.to_dict(), cwd=a.get_cwd(), schema=a.get_schema())
            a.deep_update(b, full_merge=True)
            return a

        def _ctx_for_ctx_update(ctx_for_ctx: XConfig, new_ctxs: t.Sequence[XConfig]):
            for newc in new_ctxs:
                ctx_for_ctx = _sum(newc, ctx_for_ctx)
            return ctx_for_ctx

        # use the command line context to process the other contexts
        effective_ctx = []
        ctx_for_ctx = base_ctx[-1]
        for idx, curr_ctx in enumerate(base_ctx):
            if curr_ctx.to_dict():
                # the context itself is a valid context
                curr_ctx_for_ctx = _sum(curr_ctx, ctx_for_ctx)

                if verbose > 3:
                    print_info(f"[{idx}] context to process:")
                    print_info(curr_ctx.to_dict(), pretty=True)
                    print_info(f"[{idx}] context for context")
                    print_info(curr_ctx_for_ctx.to_dict(), pretty=True)
                if verbose > 2:
                    print_info(f"[{idx}] preprocessing...")

                new_ctxs = _process_cfg_or_die(
                    curr_ctx,
                    curr_ctx_for_ctx,
                    "context",
                    run_all,
                    None,
                    True,
                    verbose > 2,
                    verbose > 3,
                )
                partial_ctx_for_ctx = _ctx_for_ctx_update(ctx_for_ctx, new_ctxs)

                if verbose > 3:
                    print_info(f"[{idx}] updated context for context")
                    print_info(partial_ctx_for_ctx.to_dict(), pretty=True)
                if verbose > 2:
                    print_info(f"[{idx}] processing self-references...")

                new_ctxs = _process_cfg_or_die(
                    curr_ctx,
                    partial_ctx_for_ctx,
                    "context",
                    run_all,
                    None,
                    True,
                    verbose > 2,
                    verbose > 3,
                )
                ctx_for_ctx = _ctx_for_ctx_update(ctx_for_ctx, new_ctxs)

                if verbose > 3:
                    print_info(f"[{idx}] final updated context for context")
                    print_info(ctx_for_ctx.to_dict(), pretty=True)

                effective_ctx.extend(new_ctxs)

        if effective_ctx:
            effective_ctx = functools.reduce(
                lambda acc, curr: _deep_update_fn(acc, curr, verbose > 3), effective_ctx
            )
        else:
            effective_ctx = XConfig()

        if verbose > 1:
            print_info("\nFinal effective context:")
            print_info(effective_ctx.to_dict(), pretty=True)

        if command in subc.AUDIT[0]:
            from dataclasses import fields

            from pipelime.choixe.visitors.processor import ChoixeProcessingError

            print_info("\n📄 CONFIGURATION AUDIT")
            for idx, c in enumerate(base_cfg):
                if len(base_cfg) > 1:
                    name = str(config[idx]) if idx < len(config) else "command line"
                    print_info(f"\n*** {name}")
                inspect_info = c.inspect()
                for field in fields(inspect_info):
                    value = getattr(inspect_info, field.name)
                    print_info(f"🔍 {field.name}:")
                    if value or isinstance(value, bool):
                        print_info(value, pretty=True, indent_guides=False)

            print_info("\n📄 EFFECTIVE CONTEXT\n")
            print_info(effective_ctx.to_dict(), pretty=True, indent_guides=False)
            print_info("")

            try:
                effective_configs = _process_all(
                    base_cfg, effective_ctx, output, run_all, False, verbose > 2
                )
            except ChoixeProcessingError as e:
                # from rich.prompt import Confirm, Prompt

                # from pipelime.cli.wizard import Wizard

                print_warning("Some variables are not defined in the context.")
                print_error(f"Invalid configuration! {e}")
                raise typer.Exit(1)

                ### SKIP FOR NOW
                # if not Confirm.ask(
                #     "Do you want to create a new context?", default=True
                # ):
                #     print_error(f"Invalid configuration! {e}")
                #     raise typer.Exit(1)
            #
            # print_info("\n📝 Please enter a value for each variable")
            # new_ctx = Wizard.context_wizard(inspect_info.variables, effective_ctx)
            #
            # print_info("Processing configuration and context...", end="")
            # effective_configs = _process_cfg_or_die(
            #     base_cfg, new_ctx, run_all, output
            # )
            # print_info(" OK")
            #
            # outfile = Prompt.ask("\n💾 Write to (leave empty to skip)")
            # if outfile:
            #     new_ctx.save_to(Path(outfile).with_suffix(".yaml"))

            cfg_size = len(effective_configs)
            pls = "s" if cfg_size != 1 else ""
            print_info(
                f"🎉 Configuration successfully processed ({cfg_size} variant{pls})."
            )

            if verbose > 2:
                print_info("\nFinal effective configurations:")
                for idx, cfg in enumerate(effective_configs):
                    print_info(f"\n*** CONFIGURATION {idx+1}/{cfg_size} ***\n")
                    print_info(cfg.to_dict(), pretty=True)

            raise typer.Exit(0)
        else:
            from pipelime.cli.pretty_print import show_spinning_status

            with show_spinning_status("Processing configuration and context..."):
                effective_configs = _process_all(
                    base_cfg, effective_ctx, output, run_all, True, verbose
                )

            cmd_name = command
            cfg_size = len(effective_configs)
            for idx, cfg in enumerate(effective_configs):
                cfg_dict = cfg.to_dict()

                if verbose > 1:
                    print_info(f"\n*** CONFIGURATION {idx+1}/{cfg_size} ***\n")
                    print_info(cfg_dict, pretty=True)

                if command in subc.EXEC[0]:
                    if len(cfg_dict) == 0:
                        print_error("No command specified.")
                        raise typer.Exit(1)
                    if len(cfg_dict) > 1:
                        print_error("Multiple commands found.")
                        print_warning(
                            "You should use the `run` command to process a dag."
                        )
                        raise typer.Exit(1)
                    cmd_name = next(iter(cfg_dict))
                    cfg_dict = next(iter(cfg_dict.values()))

                run_command(cmd_name, cfg_dict, verbose, dry_run)
    else:
        from pipelime.cli.pretty_print import print_error

        print_error("No command specified.")
        raise typer.Exit(1)


def run_command(command: str, cmd_args: t.Mapping, verbose: int, dry_run: bool):
    """
    Run a pipelime command.
    """

    import time
    from pydantic.error_wrappers import ValidationError

    from pipelime.cli.pretty_print import print_info, print_command_outputs
    from pipelime.cli.utils import get_pipelime_command_cls, time_to_str, show_field_alias_valerr

    try:
        cmd_cls = get_pipelime_command_cls(command)
    except ValueError:
        raise typer.Exit(1)

    if verbose > 2:
        print_info(f"\nCreating command `{command}` with options:")
        print_info(cmd_args, pretty=True)

    try:
        cmd_obj = cmd_cls(**cmd_args)
    except ValidationError as e:
        show_field_alias_valerr(e)
        raise e

    if verbose > 0:
        print_info(f"\nCreated command `{command}`:")
        print_info(cmd_obj.dict(), pretty=True)

    if dry_run or verbose > 0:
        print_info(f"\nRunning `{command}`...")

    start_time = time.perf_counter_ns()
    if not dry_run:
        cmd_obj()
    end_time = time.perf_counter_ns()
    print_info("\nCommand executed in " + time_to_str(end_time - start_time))

    print_info(f"\n`{command}` outputs:")
    print_command_outputs(cmd_obj)


def _create_typer_app(
    *,
    app_name: str = "Pipelime",
    entry_point: t.Optional[str] = None,
    app_description: t.Optional[str] = None,
    version: t.Optional[str] = None,
    extra_args: t.Sequence[str] = list(),
):
    import sys

    if entry_point is None and len(sys.argv) > 0:
        entry_point = Path(sys.argv[0]).name

    # if there is no other args, run the app with no args
    if extra_args and len(sys.argv) > 1:
        sys.argv.extend(extra_args)

    def _preproc_docs(func):
        desc = (
            app_description
            if app_description
            else f"{app_name} Command Line Interface."
        )
        n_appname = ("n " if app_name[0] in "aeiouAEIOU" else " ") + app_name
        if desc[-1] != ".":
            desc += "."
        func.__doc__ = func.__doc__.format(
            desc, entry_point or app_name.casefold(), n_appname
        )
        return func

    if version is not None:
        VersionCallback.get_version = lambda: version

    app = typer.Typer(pretty_exceptions_enable=False)
    app.command(
        add_help_option=False,
        no_args_is_help=True,
        context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    )(_preproc_docs(pl_main))
    return app


def run_typer_app(
    *,
    app_name: str = "Pipelime",
    entry_point: t.Optional[str] = None,
    app_description: t.Optional[str] = None,
    version: t.Optional[str] = None,
    extra_args: t.Sequence[str] = list(),
):
    app = _create_typer_app(
        app_name=app_name,
        entry_point=entry_point,
        app_description=app_description,
        version=version,
        extra_args=extra_args,
    )
    app()


if __name__ == "__main__":
    run_typer_app()



import typing as t
import trimesh
from trimesh.parent import Geometry

from pipelime.items import Item
from pipelime.items.base import deferred_classattr


class Model3DItem(Item[Geometry]):
    """Base class for all mesh types. Subclasses should implement `file_extensions`,
    and, optionally, `save_options` and `load_options`.
    """

    @deferred_classattr
    def default_concrete(cls):
        return OBJModel3DItem

    @classmethod
    def save_options(cls) -> t.Mapping[str, t.Any]:
        """Subclass can override and return specific saving options.

        Returns:
          t.Mapping[str, t.Any]: saving parameters
        """
        return {}

    @classmethod
    def load_options(cls) -> t.Mapping[str, t.Any]:
        """Subclass can override and return specific loading options.

        Returns:
          t.Mapping[str, t.Any]: loading parameters
        """
        return {}

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> Geometry:
        return trimesh.load(
            fp, file_type=cls.file_extensions()[0][1:], **cls.load_options()  # type: ignore
        )

    @classmethod
    def encode(cls, value: Geometry, fp: t.BinaryIO):
        value.export(fp, file_type=cls.file_extensions()[0][1:], **cls.save_options())


class STLModel3DItem(Model3DItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".stl",)


class OBJModel3DItem(Model3DItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".obj",)


class PLYModel3DItem(Model3DItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".ply",)


class OFFModel3DItem(Model3DItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".off",)


class GLBModel3DItem(Model3DItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".glb",)



import typing as t
import pydantic as pyd

import pipelime.sequences as pls
from pipelime.sequences.pipes import PipedSequenceBase
from pipelime.stages import StageInput
from pipelime.utils.pydantic_types import CallableDef


@pls.piped_sequence
class MappedSequence(PipedSequenceBase, title="map"):
    """Applies a stage on all samples."""

    stage: StageInput = pyd.Field(...)

    def __init__(self, stage: StageInput, **data):
        super().__init__(stage=stage, **data)  # type: ignore

    def get_sample(self, idx: int) -> pls.Sample:
        return self.stage(self.source[idx])


@pls.piped_sequence
class ConditionallyMappedSequence(PipedSequenceBase, title="map_if"):
    """Applies a stage only when a condition returns True."""

    stage: StageInput = pyd.Field(...)
    condition: CallableDef = pyd.Field(
        ...,
        description=(
            "A callable that returns True if the sample should be mapped through "
            "the stage. Accepted signatures:\n"
            "  `() -> bool`\n"
            "  `(index: int) -> bool`\n"
            "  `(index: int, sample: Sample) -> bool`\n"
            "  `(index: int, sample: Sample, source: SamplesSequence) -> bool`."
        ),
    )

    # This convoluted implementation is needed to support pickling and multiprocessing
    _call_condition = pyd.PrivateAttr()

    @staticmethod
    def _call0(fn: t.Callable[[], bool], *args, **kwargs) -> bool:
        return fn()

    @staticmethod
    def _call1(fn: t.Callable[[int], bool], idx: int, *args, **kwargs) -> bool:
        return fn(idx)

    @staticmethod
    def _call2(
        fn: t.Callable[[int, pls.Sample], bool],
        idx: int,
        x: pls.Sample,
        *args,
        **kwargs,
    ) -> bool:
        return fn(idx, x)

    @staticmethod
    def _call3(
        fn: t.Callable[[int, pls.Sample, pls.SamplesSequence], bool],
        idx: int,
        x: pls.Sample,
        s: pls.SamplesSequence,
        *args,
        **kwargs,
    ) -> bool:
        return fn(idx, x, s)

    def __init__(self, **data):
        from inspect import signature, Parameter

        super().__init__(**data)

        num_prms = len(self.condition.args)
        has_var_pos = self.condition.has_var_positional
        if num_prms > 4 or (num_prms == 4 and not has_var_pos):
            raise ValueError(
                f"Invalid signature for `condition`: {self.condition.full_signature}"
            )

        if num_prms == 3 or has_var_pos:
            self._call_condition = self._call3
        elif num_prms == 2:
            self._call_condition = self._call2
        elif num_prms == 1:
            self._call_condition = self._call1
        else:
            self._call_condition = self._call0

    def get_sample(self, idx: int) -> pls.Sample:
        x = self.source[idx]
        if self._call_condition(self.condition, idx, x, self.source):  # type: ignore
            return self.stage(x)
        return x


class MappingConditionProbability(pyd.BaseModel):
    """A condition that returns True with a given probability."""

    probability: float = pyd.Field(
        ...,
        alias="p",
        ge=0.0,
        le=1.0,
        description="Probability that the sample will be mapped through the stage.",
    )
    seed: t.Optional[int] = pyd.Field(None, description="Random seed.")

    _generator = pyd.PrivateAttr(None)

    def __call__(self) -> bool:
        # create the random generator here so that every process has its own
        if self._generator is None:
            import random

            self._generator = random.Random(self.seed)
        return self._generator.random() < self.probability


class MappingConditionIndexRange(pyd.BaseModel):
    """A condition that returns True if the sample index is in a given range.
    Negative indices are supported.
    """

    start: int = pyd.Field(0, description="First index (inclusive).")
    stop: t.Optional[int] = pyd.Field(None, description="Last index (exclusive).")
    step: int = pyd.Field(1, description="Step.")

    def __call__(self, idx: int, _: pls.Sample, seq: pls.SamplesSequence) -> bool:
        start = (len(seq) + self.start) if self.start < 0 else self.start
        stop = (
            len(seq)
            if self.stop is None
            else ((len(seq) + self.stop) if self.stop < 0 else self.stop)
        )
        return idx in range(start, stop, self.step)



import typing as t
from pydantic import BaseModel
from pydantic.typing import display_as_type
import pipelime.cli.pretty_print as ppp

if t.TYPE_CHECKING:
    from pipelime.choixe import XConfig


class InvalidValue(Exception):
    pass


class ColoredPath:
    COLORS = [
        "[#FF7F00]",
        "[#33A02C]",
        "[#6A3D9A]",
        "[#E31A1C]",
        "[#B15928]",
        "[#1F78B4]",
        "[#FDBF6F]",
        "[#B2DF8A]",
        "[#CAB2D6]",
        "[#FB9A99]",
        "[#FFFF99]",
        "[#A6CEE3]",
    ]

    def __init__(self, path: str = "", cidx: int = -1):
        self.path = path
        self.cidx = cidx

    def as_map(self, path: str):
        cidx = (self.cidx + 1) % len(self.COLORS)
        col_hex = self.COLORS[cidx]
        return ColoredPath(
            ".".join(filter(None, [self.path, col_hex + path + "[/]"])), cidx=cidx
        )

    def as_list(self, path: str):
        cidx = (self.cidx + 1) % len(self.COLORS)
        col_hex = self.COLORS[cidx]
        return ColoredPath(
            self.path + col_hex + "\[" + path + "][/]", cidx=cidx  # type: ignore # noqa: W605
        )

    def __str__(self) -> str:
        return str(self.path)

    def __repr__(self) -> str:
        return repr(self.path)


class FieldModel:
    def __init__(
        self,
        name: str,
        outer_type: t.Type,
        description: t.Optional[str],
        is_required: bool,
        default_value: t.Any,
        is_model: bool,
    ):
        self.name = name
        self.outer_type = outer_type
        self.description = description
        self.is_required = is_required
        self.default_value = default_value
        self.is_model = is_model


class Wizard:
    @staticmethod
    def help_message(title: str):
        ppp.print_info("\n🪄  " + title)
        ppp.print_info("- `\"` or `'` to enforce string values")
        ppp.print_info(
            "- `[` and `{` followed by a new-line to start a list or dict "
            "(one item per line, use `:` to separate key and value)"
        )
        ppp.print_info("- `]` and `}` followed by a new-line to end a list or dict")
        ppp.print_info(
            "- `< \[model]` to begin a wizard configuration for a pydantic model "  # type: ignore # noqa: W605
            "(should be explicitly listed in the type list)."
        )
        ppp.print_info(
            "- `? \[class.path]` to begin a wizard configuration "  # type: ignore # noqa: W605
            "for a choixe `$call` directive."
        )
        ppp.print_info("- `! \[class.path]` to add a choixe `$symbol` directive.")  # type: ignore # noqa: W605
        ppp.print_info(
            "- `# \[name]` to begin a wizard configuration "  # type: ignore # noqa: W605
            "for a pipelime command, stage or operation."
        )
        ppp.print_info(
            "- `c# \[name]`, `s# \[name]`, `o# \[name]` as above, "  # type: ignore # noqa: W605
            "but specifying the type."
        )

    @staticmethod
    def print_yaml(cfg):
        import yaml

        ppp.print_info("\n✨ FINAL YAML")
        ppp.print_info("==============")
        ppp.print_info(yaml.safe_dump(cfg, sort_keys=False))

    @staticmethod
    def model_cfg_wizard(model_cls: t.Union[t.Type[BaseModel], str]):
        Wizard.help_message("Configuration Wizard")

        try:
            if isinstance(model_cls, str):
                model_cls = _get_pipelime_type(model_cls, True)  # type: ignore
        except InvalidValue:
            return

        cfg = _iterate_model_fields(model_cls)  # type: ignore
        Wizard.print_yaml(cfg)
        return cfg

    @staticmethod
    def context_wizard(variables: t.Mapping, default_ctx: "XConfig"):
        from pydash import set_
        from pipelime.choixe import XConfig

        Wizard.help_message("Context Wizard")

        new_ctx = {}
        for var, val in variables.items():
            default_value = default_ctx.deep_get(var, default=...)
            if default_value is ...:
                if val is not None:
                    default_value = val
            val = _get_field_value(
                FieldModel(
                    name=var,
                    outer_type=t.Any if default_value is ... else type(default_value),
                    description=None,
                    is_required=default_value is ...,
                    default_value=default_value,
                    is_model=False,
                ),
                prefix=ColoredPath(),
            )
            set_(new_ctx, var, val)
        Wizard.print_yaml(new_ctx)
        return XConfig(new_ctx)


def _get_pipelime_type(
    name: str, is_command: bool = True, is_stage: bool = True, is_operation: bool = True
) -> t.Type[BaseModel]:
    from pipelime.cli.utils import PipelimeSymbolsHelper

    cmd_cls = None
    if is_command:
        cmd_cls = PipelimeSymbolsHelper.get_command(name)
    if cmd_cls is None and is_stage:
        cmd_cls = PipelimeSymbolsHelper.get_stage(name)
    if cmd_cls is None and is_operation:
        cmd_cls = PipelimeSymbolsHelper.get_operator(name)

    if cmd_cls is None:
        PipelimeSymbolsHelper.show_error_and_help(
            name,
            should_be_cmd=is_command,
            should_be_op=is_operation,
            should_be_stage=is_stage,
        )
        raise InvalidValue()

    return cmd_cls[1]


def _get_value_list(prefix: ColoredPath):
    from rich.prompt import Prompt

    value_list = []
    while True:
        value = Prompt.ask(f"{prefix}[{len(value_list)}]")
        if value == "]":
            return value_list
        try:
            value_list.append(
                _decode_value(value, prefix.as_list(str(len(value_list))))
            )
        except InvalidValue:
            pass


def _get_value_map(prefix: ColoredPath):
    from rich.prompt import Prompt

    value_map = {}
    while True:
        key_val = Prompt.ask(f"{prefix}" "{}")
        if key_val == "}":
            return value_map
        key, _, value = key_val.partition(":")
        try:
            value_map[key] = _decode_value(value, prefix.as_map(key))
        except InvalidValue:
            pass


def _get_general_field_value(prefix: ColoredPath, default=...):
    from rich.prompt import Prompt

    while True:
        value = Prompt.ask("Enter value", default=default)
        try:
            return _decode_value(value, prefix)
        except InvalidValue:
            pass


def _get_model(v: str, prefix: ColoredPath):
    from pipelime.choixe.utils.imports import import_symbol

    if not v:
        raise InvalidValue()
    try:
        model_cls = import_symbol(v)
    except ImportError:
        ppp.print_error(f"Cannot import model class {v}")
        raise InvalidValue()
    try:
        return _iterate_model_fields(model_cls, prefix)
    except TypeError as e:
        ppp.print_error(f"Invalid symbol: {v} ({e})")
        raise InvalidValue()


def _get_callable(v: str, prefix: ColoredPath):
    from pipelime.choixe.utils.imports import import_symbol

    if not v:
        raise InvalidValue()
    try:
        clb_type = import_symbol(v)
    except ImportError:
        ppp.print_error(f"Cannot import callable {v}")
        raise InvalidValue()
    try:
        args = (
            _iterate_model_fields(clb_type, prefix)
            if ppp._is_model(clb_type)
            else _iterate_callable_args(clb_type, prefix)
        )
        return {"$call": v, "$args": args}
    except TypeError as e:
        ppp.print_error(f"Invalid symbol: {v} ({e})")
        raise InvalidValue()


def _get_symbol(v: str):
    from pipelime.choixe.utils.imports import import_symbol

    if not v:
        raise InvalidValue()
    try:
        _ = import_symbol(v)
    except ImportError:
        ppp.print_warning(
            f"{v} {ppp._short_line()} Symbol will be included, but importing has failed"
        )
    return f"$symbol({v})"


def _get_pipelime_object(value: str, prefix: ColoredPath):
    v = value[1:].strip() if value[0] == "#" else value[2:].strip()
    if not v:
        raise InvalidValue()

    pl_type = _get_pipelime_type(
        v,
        is_command=value[0] in ("#", "c"),
        is_stage=value[0] in ("#", "s"),
        is_operation=value[0] in ("#", "o"),
    )
    args = _iterate_model_fields(pl_type, prefix)
    return {v: args}


def _decode_value(value, prefix: ColoredPath):
    value = value.strip()
    if len(value) > 1 and value[0] in ("'", '"') and value[-1] == value[0]:
        return value[1:-1]
    if value == "[":
        return _get_value_list(prefix)
    if value == "{":
        return _get_value_map(prefix)
    if value[0] == "<":
        return _get_model(value[1:].strip(), prefix)
    if value[0] == "?":
        return _get_callable(value[1:].strip(), prefix)
    if value[0] == "!":
        return _get_symbol(value[1:].strip())
    if value[0] == "#" or (len(value) > 1 and value[1] == "#"):
        return _get_pipelime_object(value, prefix)
    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False
    if value.lower() in (None, "", "none", "null", "nul"):
        return None
    try:
        num = int(value)
        return num
    except ValueError:
        pass
    try:
        num = float(value)
        return num
    except ValueError:
        pass

    return value


def _get_field_value(field: FieldModel, prefix: ColoredPath):
    from rich.table import Table
    from rich import box
    from rich.prompt import Confirm
    from rich.pretty import Pretty
    from rich import print as rprint

    field_prefix = prefix.as_map(field.name)
    header = [
        f"{field_prefix}\n«"
        + display_as_type(field.outer_type).replace("[", r"\[")  # noqa: W605
        + "»"
    ]
    if not field.is_required:
        header.append("Default")

    desc = field.description if field.description else ""

    table = Table(
        *header,
        box=box.SIMPLE_HEAVY,
        show_header=True,
        show_footer=False,
        show_lines=True,
        width=80,
    )

    if field.is_required:
        table.add_row(desc)
        rprint(table)
        if field.is_model:
            return _get_general_field_value(
                prefix=field_prefix,
                default=f"< {field.outer_type.__module__}.{field.outer_type.__name__}",
            )
        return _get_general_field_value(prefix=field_prefix)
    else:
        default_value = field.default_value
        if field.is_model:
            if default_value is not None:
                default_value = default_value.dict()
                table.add_row(
                    desc, Pretty(default_value, indent_guides=True, expand_all=True)
                )
            else:
                table.add_row(desc, "None")

            rprint(table)
            if Confirm.ask("Accept default?", default=True):
                return default_value
            return _get_general_field_value(
                prefix=field_prefix,
                default=f"< {field.outer_type.__module__}.{field.outer_type.__name__}",
            )

        table.add_row(desc, Pretty(default_value, indent_guides=True, expand_all=True))
        rprint(table)
        return _get_general_field_value(prefix=field_prefix, default=str(default_value))


def _iterate_model_fields(model_cls: t.Type[BaseModel], prefix=ColoredPath()):
    if not ppp._is_model(model_cls):
        raise TypeError("not a pydantic model")

    cfg = {}
    for field in model_cls.__fields__.values():
        if not field.field_info.exclude:
            is_model = ppp._is_model(field.outer_type_)
            has_root_item = (
                ("__root__" in field.outer_type_.__fields__) if is_model else False
            )
            field_outer_type = (
                field.outer_type_.__fields__["__root__"].outer_type_
                if has_root_item
                else field.outer_type_
            )

            try:
                cfg[field.name] = _get_field_value(
                    FieldModel(
                        name=field.name,
                        outer_type=field_outer_type,
                        description=field.field_info.description,
                        is_required=bool(field.required),
                        default_value=field.get_default(),
                        is_model=is_model,
                    ),
                    prefix,
                )
            except InvalidValue:
                pass
    return cfg


def _iterate_callable_args(clb_type, prefix=ColoredPath()):
    import inspect

    cfg = {}
    for field in inspect.signature(
        clb_type.__init__ if inspect.isclass(clb_type) else clb_type
    ).parameters.values():
        if field.name in ("self", "cls"):
            continue
        if field.kind in (
            inspect.Parameter.VAR_POSITIONAL,
            inspect.Parameter.POSITIONAL_ONLY,
        ):
            raise TypeError(
                f"{inspect.Parameter.VAR_POSITIONAL.description} and "
                f"{inspect.Parameter.POSITIONAL_ONLY.description} parameters "
                "are not supported"
            )

        try:
            cfg[field.name] = _get_field_value(
                FieldModel(
                    name=field.name,
                    outer_type=t.Any
                    if field.annotation is inspect.Parameter.empty
                    else field.annotation,
                    description=None,
                    is_required=field.default is inspect.Parameter.empty,
                    default_value=field.default,
                    is_model=False,
                ),
                prefix,
            )
        except InvalidValue:
            pass
    return cfg



from pipelime.commands.general import (
    TimeItCommand,
    PipeCommand,
    CloneCommand,
    ConcatCommand,
    ZipCommand,
    AddRemoteCommand,
    RemoveRemoteCommand,
    ValidateCommand,
    MapCommand,
    SortCommand,
    FilterCommand,
)
from pipelime.commands.shell import ShellCommand
from pipelime.commands.split_ops import (
    SplitByQueryCommand,
    SplitByValueCommand,
    SplitCommand,
)
from pipelime.commands.toy_dataset import ToyDatasetCommand
from pipelime.commands.piper import RunCommand, DrawCommand, WatchCommand



from abc import ABC, abstractmethod
from enum import Enum
import typing as t

from pydantic import BaseModel, Field, PrivateAttr

from pipelime.piper.progress.tracker.base import Tracker, TrackedTask, TqdmTask


# The return type is a hack to fool the type checker
# otherwise it would complain when calling the command
# TODO: forward with typing.ParamSpec (New in Python 3.10) to get full type checking
@t.overload
def command(
    *,
    title: t.Optional[str] = None,
    **__config_kwargs,
) -> t.Callable[[t.Callable[..., None]], t.Callable[..., t.Type["PipelimeCommand"]]]:
    ...


# The return type is a hack to fool the type checker
# otherwise it would complain when calling the command
# TODO: forward with typing.ParamSpec (New in Python 3.10) to get full type checking
@t.overload
def command(
    __func: t.Callable[..., None]
) -> t.Callable[..., t.Type["PipelimeCommand"]]:
    ...


def command(__func=None, *, title: t.Optional[str] = None, **__config_kwargs):
    """Creates a full-fledged PipelimeCommand from a general function.
    The command will have the same exact signature and docstring of the function.
    Field names and types will taken from the function parameters and validated as usual.
    Any function signature is allowed, including positional-only, keyword-only,
    variable positional, variable keyword parameters. Variable positional and keyword
    can be annotated to get type checking and validation.

    You are encouraged to use pydantic.Field as default value to further specify any
    field properties, such as `default_factory`, `alias`, `description`, etc, which
    are used to show meaningful help messages through `pipelime help`.
    Also, call it with extra `__config_kwargs` to add any pydantic config parameters,
    such as `title`, `arbitrary_types_allowed`, etc.

    Examples:
        Create a command with the same name of the function::

            @command
            def addup(a: int, b: int):
                print(a + b)

        then you can use it as::

            $ pipelime addup +a 1 +b 2

        Create a command with custom name and non-pydantic types::

            @command(title="add-up-these", arbitrary_types_allowed=True)
            def my_addup_func(a: int, b: int, c: t.Optional[MyType] = None):
                print(a + b)

        then you can use it as::

            $ pipelime add-up-these +a 1 +b 2

        Create a command with field aliases and default factories::

            @command
            def merge_lists(
                first: t.List[int] = Field(default_factory=list, alias="f"),
                second: t.List[int] = Field(default_factory=list, alias="s"),
            ):
                print(first + second)

        then you can use it as::

            $ pipelime merge_lists +f 1 +f 2 +f 3 +s 4 +s 5 +s 6

        Create a command with help messages::

            @command(title="merge")
            def merge_lists(
                first: t.List[int] = Field(
                    default_factory=list, alias="f", description="first list"
                ),
                second: t.List[int] = Field(
                    default_factory=list, alias="s", description="second list"
                ),
            ):
                '''Merge and print two lists.'''
                print(first + second)

        then show the command help::

            $ pipelime merge help
    """

    if title is not None:
        __config_kwargs["title"] = title

    def _make_cmd(func):
        import inspect
        from pydantic.fields import FieldInfo, Undefined

        def _make_field(p: inspect.Parameter):
            """Returns a tuple of (annotation, default) for a given parameter.
            NB: *args translates to a Sequence and **kwargs translates to a Mapping.
            """
            value = Field(...) if p.default is inspect.Parameter.empty else p.default

            if p.kind is p.VAR_POSITIONAL:
                ann = (
                    t.Sequence
                    if p.annotation is inspect.Signature.empty
                    else t.Sequence[p.annotation]
                )
            elif p.kind is p.VAR_KEYWORD:
                ann = (
                    t.Mapping
                    if p.annotation is inspect.Signature.empty
                    else t.Mapping[str, p.annotation]
                )
            else:
                ann = p.annotation

            return (ann, value)

        # Translates signature to pydantic fields
        # and gathers positional arguments
        fsig = inspect.signature(func)
        fields = {n: _make_field(p) for n, p in fsig.parameters.items()}
        posonly_names = [
            p.name for p in fsig.parameters.values() if p.kind is p.POSITIONAL_ONLY
        ]
        poskw_names = [
            p.name
            for p in fsig.parameters.values()
            if p.kind is p.POSITIONAL_OR_KEYWORD
        ]
        kwonly_names = [
            p.name for p in fsig.parameters.values() if p.kind is p.KEYWORD_ONLY
        ]
        try:
            varpos_name = next(
                p.name for p in fsig.parameters.values() if p.kind is p.VAR_POSITIONAL
            )
        except StopIteration:
            varpos_name = ""
        try:
            varkw_name = next(
                p.name for p in fsig.parameters.values() if p.kind is p.VAR_KEYWORD
            )
        except StopIteration:
            varkw_name = ""

        # set title to function name, if not specified
        __config_kwargs.setdefault("title", func.__name__)

        class _FnModel(PipelimeCommand):
            def __init__(self, *args, **kwargs):
                # positional-only arguments are not allowed in kwargs
                for p in posonly_names + [varpos_name]:
                    if p in kwargs:
                        raise TypeError(
                            f"Argument '{p}' in {self.command_name} is positional-only"
                        )

                # positional arguments must be less than declared ones
                if not varpos_name and len(args) > len(posonly_names + poskw_names):
                    raise TypeError(
                        f"{self.command_name} takes {len(posonly_names + poskw_names)} "
                        f"positional arguments but {len(args)} were given"
                    )

                # move positional arguments to data
                data = {}
                for name, value in zip(posonly_names, args):
                    data[name] = value
                for name, value in zip(poskw_names, args[len(posonly_names) :]):
                    if name in kwargs:
                        raise TypeError(
                            f"{self.command_name} got multiple values "
                            f"for argument '{name}'"
                        )
                    data[name] = value
                if varpos_name:
                    data[varpos_name] = tuple(
                        args[len(posonly_names) + len(poskw_names) :]
                    )

                # move keyword arguments to data
                for name in poskw_names + kwonly_names:
                    if name in kwargs:
                        data[name] = kwargs.pop(name)
                if varkw_name:
                    data[varkw_name] = kwargs
                    kwargs = {}

                # kwargs should be passed (user may have set extra="allow")
                super(_FnModel, self).__init__(**data, **kwargs)

            def run(self):
                # get all arguments in the right order
                func(
                    *[getattr(self, n) for n in posonly_names + poskw_names],
                    *getattr(self, varpos_name, tuple()),
                    **self.dict(include=set(kwonly_names + [varkw_name])),
                )

        # override base docstring with a custom description
        _FnModel.__doc__ = (
            f"Autogenerated Pipelime command from `{func.__module__}.{func.__name__}`."
        )

        # create the pipelime command class
        fmodel = type(
            func.__name__.replace("_", " ").title().replace(" ", "") + "Command",
            (_FnModel,),
            {
                "__module__": func.__module__,
                "__doc__": func.__doc__,
                "__annotations__": {
                    n: f[0]
                    for n, f in fields.items()
                    if f[0] is not inspect.Signature.empty
                },
                **{n: f[1] for n, f in fields.items()},
            },
            **__config_kwargs,
        )

        # override pydantic signature
        # --> we need this to pretty print the command help
        def _unwrap_default(value):
            if isinstance(value, FieldInfo):
                value = (
                    value.default
                    if value.default_factory is None
                    else value.default_factory()
                )
                if value in (Ellipsis, Undefined):
                    return inspect.Parameter.empty
            return value

        fmodel.__signature__ = fsig.replace(
            parameters=[
                p.replace(default=_unwrap_default(p.default))
                for p in fsig.parameters.values()
            ],
            return_annotation=inspect.Signature.empty,
        )
        return fmodel

    if __func is None:
        return _make_cmd
    return _make_cmd(__func)


def self_() -> "PipelimeCommand":
    from pipelime.utils.inspection import MyCaller

    parent = MyCaller(1).locals.get("self", None)
    if parent is None or not isinstance(parent, PipelimeCommand):
        raise TypeError(
            "self_() can only be used inside a function decorated as @command."
        )
    return parent


class PiperPortType(Enum):
    INPUT = "input"
    OUTPUT = "output"
    PARAMETER = "parameter"


class PiperInfo(BaseModel, extra="forbid", copy_on_model_validation="none"):
    token: str = Field("", description="The piper execution token.")
    node: str = Field("", description="The piper dag's node name.")

    @property
    def active(self) -> bool:
        return bool(self.token)


class PipelimeCommand(
    BaseModel,
    ABC,
    allow_population_by_field_name=True,
    extra="forbid",
    copy_on_model_validation="none",
):
    """Base class for all pipelime commands.
    Subclasses should implement the run method.
    """

    _piper: PiperInfo = PrivateAttr(default_factory=PiperInfo)  # type: ignore
    _tracker: t.Optional["Tracker"] = PrivateAttr(None)

    @abstractmethod
    def run(self) -> None:
        pass

    @classmethod
    def _filter_fields_by_flag(cls, flag: str, value: t.Any) -> t.Iterable[str]:
        for k, v in cls.__fields__.items():
            if v.field_info.extra.get(flag, object()) == value:
                yield k

    def _get_fields_by_flag(self, flag: str, value: t.Any) -> t.Dict[str, t.Any]:
        return {k: getattr(self, k) for k in self._filter_fields_by_flag(flag, value)}

    def get_inputs(self) -> t.Dict[str, t.Any]:
        return self._get_fields_by_flag("piper_port", PiperPortType.INPUT)

    def get_outputs(self) -> t.Dict[str, t.Any]:
        return self._get_fields_by_flag("piper_port", PiperPortType.OUTPUT)

    def _get_piper_tracker(self) -> "Tracker":
        if self._tracker is None:  # pragma: no branch
            from pipelime.piper.progress.tracker.base import Tracker
            from pipelime.piper.progress.tracker.factory import TrackCallbackFactory

            cb = TrackCallbackFactory.get_callback()
            self._tracker = Tracker(self._piper.token, self._piper.node, cb)

        return self._tracker

    @classmethod
    def command_title(cls) -> str:
        if cls.__config__.title:
            return cls.__config__.title
        return cls.__name__

    @property
    def command_name(self) -> str:
        return self.command_title()

    def set_piper_info(
        self,
        *,
        token: t.Optional[str] = None,
        node: t.Optional[str] = None,
    ):
        if token is not None:
            self._piper.token = token
        if node is not None:
            self._piper.node = node

    def track(
        self,
        seq: t.Union[t.Sequence, t.Iterable],
        *,
        size: t.Optional[int] = None,
        message: str = "",
    ) -> t.Iterable:
        """Track a sequence with a progress bar or a piper task."""

        total = len(seq) if size is None else size  # type: ignore
        with self.create_task(total, message) as t:
            yield from t.track(seq)

    def create_task(self, total: int, message: str = "") -> "TrackedTask":
        """Explicit piper task creation."""
        if self._piper.active:
            return self._get_piper_tracker().create_task(total, message)
        return TqdmTask(total, message)

    def __call__(self) -> None:
        self.run()


class NodesDefinition(BaseModel, extra="forbid", copy_on_model_validation="none"):
    """A simple interface to parse a DAG node configuration."""

    __root__: t.Mapping[str, PipelimeCommand]

    @classmethod
    def create(
        cls,
        value: t.Union[
            "NodesDefinition",
            t.Mapping[
                str,
                t.Union[
                    t.Mapping[str, t.Optional[t.Mapping[str, t.Any]]], "PipelimeCommand"
                ],
            ],
        ],
    ):
        return cls.validate(value)

    @property
    def value(self):
        return self.__root__

    def _iter(self, *args, **kwargs):
        for k, v in super()._iter(*args, **kwargs):
            # NB: `v` is the dict of params of the actual pipelime commands
            assert k == "__root__"
            assert isinstance(v, t.Mapping)
            yield k, {
                node_name: {self.__root__[node_name].command_title(): cmd_args}
                for node_name, cmd_args in v.items()
            }

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        value: t.Union[
            "NodesDefinition",
            t.Mapping[
                str,
                t.Union[
                    t.Mapping[str, t.Optional[t.Mapping[str, t.Any]]], "PipelimeCommand"
                ],
            ],
        ],
    ):
        from pydantic import ValidationError
        from pipelime.cli.utils import get_pipelime_command, show_field_alias_valerr

        if isinstance(value, NodesDefinition):
            return value
        try:
            return cls(
                __root__={
                    name: get_pipelime_command(cmd) for name, cmd in value.items()
                }
            )
        except ValidationError as e:
            show_field_alias_valerr(e)
            raise e


class DAGModel(BaseModel, extra="forbid", copy_on_model_validation="none"):
    """A Piper DAG as a `<node>: <command>` mapping."""

    nodes: NodesDefinition

    def purged_dict(self):
        from pipelime.choixe import XConfig

        return XConfig(data={"nodes": self.nodes.value}).decode()
